library(httr)
library(jsonlite)
library(dplyr)
library(tidyr)
library(stringr)

USER_AGENT <- "LSE DS202W assignment by "
ACCESS_TOKEN_ENDPOINT <- "https://www.reddit.com/api/v1/access_token"
BASE_ENDPOINT <- "https://oauth.reddit.com"

#' Authenticates the user with Reddit API using provided credentials.
#'
#' @param credentials_json_path The file path to the JSON file containing the credentials.
#' @return The headers object containing the authentication information.
#' @examples
#' authenticate("/path/to/credentials.json")
authenticate <- function(credentials_json_path) {
  # Load the credentials
  credentials <- fromJSON(credentials_json_path)

  # Create the client authentication
  client_auth <- httr::authenticate(
    credentials$app_client_id,
    credentials$app_client_secret
  )

  # Create the post data
  post_data <- list(
    grant_type = "password",
    username = credentials$reddit_username,
    password = credentials$reddit_password
  )

  # Create the headers
  user_agent <- paste0(USER_AGENT, username = credentials$reddit_username)
  headers <- httr::add_headers(`User-Agent` = user_agent)

  # Make the request
  response <- httr::POST(
    url = ACCESS_TOKEN_ENDPOINT,
    config = client_auth,
    body = post_data,
    encode = "form",
    headers = headers
  )

  # Extract the Reddit token
  reddit_token <- httr::content(response)$access_token

  # Add the Authorization to the headers
  bearer <- paste0("bearer ", reddit_token)
  headers <- httr::add_headers(
    `User-Agent` = user_agent,
    Authorization = bearer
  )

  return(headers)
}


#' Normalizes the page data
#'
#' This function takes a page object and normalizes the data by extracting the
#' necessary information and converting it to a data frame.
#'
#' @param page The page object containing the data
#'
#' @return A data frame with the normalized data
#'
#' @examples
#' page <- get_page_data()
#' normalized_data <- normalise_page(page)
normalise_page <- function(page) {
  # Extract the data
  data <- page$data$children

  output <-
    if (length(data) == 0) {
      return(tibble())
    } else {
      tibble(data, .name_repair = ~ c("data")) %>%
        unnest_wider("data") %>%
        unnest_wider("data")
    }

  # Return the data
  return(output)
}


#' Send a request to the Reddit API to retrieve a listing
#'
#' This function sends a request to the specified endpoint of the Reddit API
#' and returns the response as a listing.
#'
#' @param endpoint The endpoint of the Reddit API to send the request to.
#' @param headers A list of headers to include in the request.
#' @param params Optional parameters to include in the request.
#'
#' @return The response from the Reddit API as a listing.
#'
#' @examples
#' request_listing("https://api.reddit.com/r/news", list("User-Agent" = "MyApp"))
#'
request_listing <- function(endpoint, headers, params = NULL) {
  # Define the full URL
  endpoint_full_url <- file.path(BASE_ENDPOINT, endpoint)

  # Make the GET request
  response <- httr::GET(endpoint_full_url, headers, query = params)

  # Parse the JSON response
  page <- httr::content(response, as = "parsed", type = "application/json")

  if (str_starts(endpoint, "comments")) {
    after <- NULL

    # Try second page by default
    page <-
      tryCatch(
        {
          normalise_page(page[[2]])
        },
        error = function(e) {
          tibble()
        }
      )

    if (nrow(page) > 0) {
      page$post_id <- str_split(endpoint, "/")[[1]][2]
    } 
  } else {
    after <- page$data$after
    page <- normalise_page(page)
  }

  return(list(page, after))
}

#' Get the top posts (right now) from Reddit's front page.
#'
#' Using the endpoint https://www.reddit.com/dev/api/#GET_top
#'
#' @param listing A string indicating the listing to retrieve.
#' @param t A string indicating the time period to consider.
#' @param n_posts An integer indicating the number of posts to retrieve.
#' @param headers A list containing the headers for the request.
#'
#' @return A tibble containing the top posts.
get_posts <- function(listing, t, n_posts, headers) {
  if (missing(n_posts) || n_posts <= 0) {
    warning("n_posts must be a positive integer. Defaulting to 100.")
    n_posts <- 100
  }

  # Divide the number of posts by 100, since the maximum number of posts per
  # request is 100
  n_requests <- n_posts %/% 100 + 1
  after <- NULL

  # There is no escaping of a for loop here, since we need the 'after'
  # parameter from the previous request to make the next one
  results <- list()
  for (i in seq_len(n_requests)) {
    params <- list(limit = 100, t = t, after = after)
    output <- request_listing(listing, headers, params)
    page <- output[[1]]
    after <- output[[2]]
    results <- append(results, list(page))
  }

  results <- bind_rows(results)
  return(results[1:n_posts, ])
}

#' Get the comments from a Reddit post.
#'
#' Using the endpoint https://www.reddit.com/dev/api/#GET_comments_{article}
#'
#' @param post_id A string indicating the id of the post to get the comments from.
#' @param headers A list containing the headers for the request.
#'
#' @return A tibble containing the comments from the post.
get_comments <- function(post_id, headers) {
  endpoint <- file.path("comments", post_id)
  output <- request_listing(endpoint, headers)
  page <- output[[1]]
  return(page)
}
